(*
	Abhinav Mishra, Addisu Petros, Vanshil Shah
*)

open Ast
open Sast

type symbol_table = { 
     		 parent    : symbol_table option; 
	 mutable functions : a_fdecl list;
     mutable variables : a_variable list; 
}

type translation_environment = {
	mutable func_return_type : t option;
					scope : symbol_table; 
}


let type_of_expr (ae : Sast.a_expr) : Sast.t = 
	match ae with
	| Sast.Literal(t, _) -> t
	| Sast.String_literal(t, _) -> t
	| Sast.Bool_literal(t, _) -> t
	| Sast.Float_literal(t, _) -> t
	| Sast.Binop(t, _, _, _) -> t
	| Sast.Id(_, _, _, t) -> t
	| Sast.Call(t, _, _) -> t
	| Sast.Basic_Op(_, _, _, _, t) -> t
	| Sast.Noexpr -> Sast.No

let formal_type_conversion = function
	Ast.Int -> Sast.Int
  | Ast.Bool -> Sast.Bool
  | Ast.String -> Sast.String
  | Ast.Float -> Sast.Float
 
let basic_type_conversion = function
	Ast.Signal -> Sast.Signal
  | Ast.Array  -> Sast.Array

let f_type_conv = function
	Some(x) -> formal_type_conversion x
  | _ -> Sast.No

let find_var name var = match var with
	Primitive({a_name=n; _}) -> n=name
  | Basic({str_name=n; _}) -> n=name

let rec find_variable (scope : symbol_table) name = 
    try 
		List.find (find_var name) scope.variables
    with Not_found ->
     	match scope.parent with 
     	  Some(parent) -> find_variable parent name 
  		| _ -> raise Not_found

let find_function (scope : symbol_table) name = 
	let rec get_global_scope scope = match scope.parent with
		| None -> scope
		| Some(parent) -> (get_global_scope parent)
	in
	let g_scope = get_global_scope scope in
	try
		List.find (fun {a_fname=s; _} -> 
					s = name) g_scope.functions
	with
	Not_found ->
		raise Not_found

let add_variable (scope : symbol_table) var = 
	scope.variables <- var :: scope.variables;
	var

let string_of_binop = function 
     Add -> "Add" 
   | Sub -> "Sub"
   | Mult -> "Mult" 
   | Div -> "Div" 
   | Exp -> "Exp" 
   | Equal -> "Equal" 
   | Neq -> "Neq" 
   | Less -> "Less" 
   | Leq -> "Leq" 
   | Greater -> "Greater"
   | Geq -> "Geq" 

let types_equal t1 t2 = 
	match t1,t2 with _, _ -> if (t1 = t2) then true else false

(*let functioncall env call = *)

let rec expr env = function 
   Ast.Literal(v) -> Sast.Literal(Sast.Int, v)
 | Ast.String_literal(v) -> Sast.String_literal(Sast.String, v)
 | Ast.Bool_literal(v) -> Sast.Bool_literal(Sast.Bool, v)
 | Ast.Float_literal(v) -> Sast.Float_literal(Sast.Float, v)
 | Ast.Binop (e1,o,e2) -> 
   ( 
	let e1 = expr env e1 
    and e2 = expr env e2 in 

	let t1 = type_of_expr e1 in
	let t2 = type_of_expr e2 in

	match o with
		(Add | Sub | Mult | Div | Exp) -> 
			if types_equal t1 t2 
				then Sast.Binop(t1, e1, o, e2)
			else 
				raise(Failure("Binop " ^ (string_of_binop o) ^ " have invalid operands."))
		
	  | (Equal | Neq | Less | Leq | Greater | Geq) ->
			if (types_equal t1 t2) 
				then Sast.Binop(Sast.Bool, e1, o, e2)
			else
				raise(Failure("Binop " ^ (string_of_binop o) ^ " have invalid operands."))
     )
  | Ast.Id(vname, exp) -> (
	let vdecl = try
		find_variable env.scope vname;
	with Not_found ->
		print_endline vname;
		raise(Failure("Variable not found"))
	in

	let ex = expr env exp in
	let ex_type = type_of_expr ex in

	let var = match vdecl with
		Primitive(x) -> x
	  | Basic(_) -> raise(Failure("Referring to signal"))
	in

	if var.a_typ <> ex_type && ex_type <> Sast.No
	then 
		raise(Failure("Assign types don't match"))
	else
		print_endline "";
	
	match var.a_typ with
		Sast.Int -> 
			Sast.Id(Sast.Var, vname, ex, Sast.Int)
	  | Sast.Float -> 
			Sast.Id(Sast.Var, vname, ex, Sast.Float)
	  | Sast.Bool ->
	  		Sast.Id(Sast.Var, vname, ex, Sast.Bool)
	  | Sast.String ->
	  		Sast.Id(Sast.Var, vname, ex, Sast.String)
	  | _ ->
	  		raise(Failure("Incorrect variable type"))

	)
  | Ast.Call(f_name, exprs) ->
  	let func_found = find_function env.scope f_name in
	let t_ret_type = func_found.a_ret_type in
	let formal_types = List.map (fun {form_type=s; _} -> formal_type_conversion s) func_found.a_formals in
	let checked_exprs = List.map (expr env) exprs in
	let checked_types = List.map type_of_expr checked_exprs in

	let rec typesMatch list1 list2 = match list1,list2 with
		| [], [] -> true
		| [], _ -> raise(Failure(" found parameters when expecting none "))
		| _ , [] -> raise(Failure(" found no parameters when expecting parameters"))
		| _ , _ -> 
			try
				( types_equal (List.hd(list1)) (List.hd(list2)) ) && 
					typesMatch (List.tl(list1)) (List.tl(list2))
			with Failure("hd") ->
				raise(Failure(" trying List.hd on [] in exprCheck:Ast.Call"))
	in

	let types_same = typesMatch formal_types checked_types in

	if types_same
	then
		match t_ret_type with
			Some(ret_val) ->
				Sast.Call(ret_val, f_name, checked_exprs)
		  | None -> 
		  		Sast.Call(Sast.No, f_name, checked_exprs)
	else
		raise(Failure("Arg types not same"))

  | Ast.Basic_Op(b_type, name, lit, id, exp) ->
  	(* If we're not in a stream function we can't do this kind of op *)
  	if env.func_return_type <> None
	then
		raise(Failure("Can't do signal/array ops in normal functions"))
	else
		(* Lets find the signal first *)
		let vdecl = try
			find_variable env.scope name
		with
			Not_found ->
				raise(Failure("Array/Signal not found"))
		in

		let a_type = basic_type_conversion b_type in
		
		let var = match vdecl with
			Primitive(_) -> raise(Failure("Basic operation on wrong type of variable"))
		  | Basic(x) ->
		  	if x.str_typ <> a_type
			then
				raise(Failure("Type of signal/array incorrect"))
			else
				x
		in
		ignore (var);
		(* Evaluate the expression *)
		let a_exp = expr env exp in

		if id <> "!Noid"
		then 
			let id_found = try
				ignore (find_variable env.scope id);
			with
				Not_found ->
					raise(Failure("ID doesn't exist"))
			in
			ignore (id_found);	
			
		(* The literal thats in Basic_Op is going to be
			dependent on the a_type

			if a_type is a signal, then lit=0 means we just have t
									lit>0 means t-lit
			if a_type is an array, then lit means Array[lit]
		*)
			Sast.Basic_Op(a_type, name, String_(id), a_exp, Sast.Float)
		else
			Sast.Basic_Op(a_type, name, Int_(lit), a_exp, Sast.Float)
		
  | Ast.Noexpr ->
	Sast.Noexpr

let rec stmt env = function
	Ast.Block(st_list) ->
		let sl = List.map (fun s -> stmt env s) st_list in
		Sast.Block(sl)
  | Ast.Sum(var, lit, i1, i2, ex) ->
	(* TODO: Make sure that i1 is less than i2 *)
	let vdecl = try 
	    ignore (find_variable env.scope var);
		raise(Failure("Identifier already exists"))
	with 
		Not_found ->  (
			let new_vdecl = Sast.Primitive({a_typ=(Sast.Int); a_name=var; a_exp=Sast.Noexpr;}) in
			ignore (add_variable env.scope new_vdecl);
		) 
	in
	vdecl;

	let new_vdecl = Sast.Primitive({a_typ=(Sast.Int); a_name=lit; a_exp=Sast.Noexpr;}) in
	ignore (add_variable env.scope new_vdecl);
	
  	let exp = expr env ex in

	let t1 = type_of_expr exp in

	Sast.Sum(t1, var, lit, i1, i2, exp)

  | Ast.Basic_Sum(s1, s2, i1, i2, exp) ->
  	let vdecl = try
		find_variable env.scope s1
	with
		Not_found -> 
			raise(Failure("Signal doesn't exit"))
	in

	let var = match vdecl with
		Primitive(x) -> raise(Failure("Incorrect variable type"))
	  | Basic(x) -> x
	in
	
	(* Var is a variable of type a_stream *)
	if var.str_typ <> Sast.Signal
	then
		raise(Failure("Array type doesn't belong here"));

	let new_vdecl = Sast.Primitive({a_typ=(Sast.Float); a_name=s2; a_exp=Sast.Noexpr;}) in
	ignore (add_variable env.scope new_vdecl);
	
  	let ex = expr env exp in

	let t1 = type_of_expr ex in

	Sast.Basic_Sum(t1, var.str_name, s2, i1, i2, ex)
  | Ast.Return(ex) ->
		(match env.func_return_type with
		| None	-> raise(Failure("Invalid return statement"))
		| Some(x) ->
			let exp = expr env ex in
			let typ = type_of_expr exp in
			if typ <> x
			then
				raise(Failure("Invalid return type"))
			else
				Sast.Return(exp)
		)
	
  | Ast.Expr(exp) ->
  	let ex =  expr env exp in

	Sast.Expr(Sast.Expression, ex)

  | Ast.If(exp, st1, st2) ->
	let exp1 = expr env exp in
	let typ = type_of_expr exp1 in

	if typ <> Sast.Bool
	then
		raise(Failure("If argument invalid"))
	else
		Sast.If(exp1, stmt env st1, stmt env st2)

  | Ast.For(e1, e2, e3, st) ->
  	let exp1 = expr env e1 in
	let exp2 = expr env e2 in
	let exp3 = expr env e3 in

	let typ2 = type_of_expr exp2 in

	if typ2 <> Sast.Bool
	then 
		raise(Failure("For loop incorrect"))
	else
		Sast.For(exp1, exp2, exp3, stmt env st)

  | Ast.While(exp, st) ->
  	let ex = expr env exp in

	let typ = type_of_expr ex in

	if typ <> Sast.Bool
	then
		raise(Failure("While condition requires boolean expression"))
	else
		Sast.While(ex, stmt env st)

  |	Ast.Prim_Assign(var) -> 
  	(* 1. Evaluate the type of rhs expression, make sure its the same as prim_type
		2. Make sure the variable name doesnt exist in this scope
	*)
	let vdecl = try 
	    ignore (find_variable env.scope var.name);
		raise(Failure("Identifier already exists"))
	with 
	Not_found ->  (
		(* TODO: add_variable returns the var *)
		let typ = formal_type_conversion var.typ in
		let a_expr = expr env var.exp in
		let type_expr = type_of_expr a_expr in

		if type_expr <> No && typ <> type_expr
		then
			raise(Failure("Type of ID doesn't match expression"))
		else (
			let new_vdecl = Sast.Primitive({a_typ=(typ); a_name=var.name; a_exp=a_expr;}) in
	  		ignore (add_variable env.scope new_vdecl);
			Sast.Prim_Assign(Sast.Var, new_vdecl)
		)
	) in
	vdecl

  | Ast.Basic_Dec(name, b_type, l_opt) ->
	(* If this is not a stream function then we can't work with signals*)
  	if env.func_return_type <> None
	then
		raise(Failure("Can't declare basic types in normal functions"))
	else
		let vdecl = try
			ignore (find_variable env.scope name);
			raise(Failure("Variable already exists"))
		with
			Not_found ->
				(* Create the signal *)
				let typ = basic_type_conversion b_type in

				if typ = Sast.Array && l_opt = -1
				then
					raise(Failure("Need size for array"));
					
				let new_vdecl = Sast.Basic({str_typ=typ; str_name=name;}) in
				ignore (add_variable env.scope new_vdecl);
				Sast.Basic_Dec(typ, name, l_opt)
		in
		vdecl
  | Ast.Print(exp) ->
  	let exp1 = expr env exp in
	let typ1 = type_of_expr exp1 in

	Sast.Print(typ1, exp1)

let check_vdecls env = function
	Ast.Variable_Dec(var) ->
		let vdecl = try 
			ignore (find_variable env.scope var.name);
			raise(Failure("Variable already exists"))
		with 
		  | Not_found -> 
			(* TODO: add_variable returns the var *)
			let typ = formal_type_conversion var.typ in
			let a_expr = expr env var.exp in
			let type_expr = type_of_expr a_expr in

			if type_expr <> No && typ <> type_expr
			then
				raise(Failure("Type of ID doesn't match expression"))
			else (
				let new_vdecl = Primitive({a_typ=(typ); a_name=var.name; a_exp=a_expr;}) in
				ignore (add_variable env.scope new_vdecl);
				Sast.Variable_Dec(Sast.Var, new_vdecl)
			)
		in
		vdecl

let create_new_env parent_env ret_type =
	let new_scope = {
		parent= Some parent_env.scope;
		functions=[];
		variables=[];
	} in
	let new_env = {
		func_return_type = ret_type;
		scope = new_scope;
	} in
	new_env

let get_global_environment =
	let new_scope = {
		parent=None;
		functions=[];
		variables=[]; 
	} in
	let new_env = {
		func_return_type=None;
		scope=new_scope;
	} in
	new_env

let variable_from_formal formal =
	let new_var = Primitive({a_typ = formal_type_conversion formal.form_type; a_name=formal.form_name; a_exp=Sast.Noexpr})  in
	new_var

let add_function globe_env fn =
	let func_exists = try
		ignore (find_function globe_env.scope fn.fname);
		raise(Failure("Function already exists"))
	with
		Not_found ->
			(* Convert the return type from
				Ast.prim option -> Sast.t option
			*)
			let new_ret_type = f_type_conv fn.ret_type in

			(* Create temporary body to fill in our function *)
			let new_body = [] in
			let new_function = {a_is_stream=fn.is_stream; 
								a_fname=fn.fname;
								a_ret_type=(match fn.is_stream with
											false -> Some new_ret_type
										  | true -> None);
								a_formals=fn.formals;
								a_locals=[];
								a_old_body=fn.body;
								a_body=new_body;} in

			(* Add the function to the global scope *)
			globe_env.scope.functions <- new_function::globe_env.scope.functions;

	in
	func_exists

let string_of_opt = function
	Some(_) -> print_endline "Has a value"
  | None -> print_endline "Is stream"

let string_of_variables = function
	Primitive(x) -> print_endline ("Primitive " ^ x.a_name)
  | Basic(x) -> print_endline ("Stream " ^ x.str_name)

let check_function_bodies globe_env fn =
	print_endline ("==============Checking function: " ^ fn.a_fname ^ "==============");

	(* Create a new environment *)
	let f_env = create_new_env globe_env fn.a_ret_type in

	(* Convert all of the formals to variables so that we can 
		add them to the scope of the current function
	*)
	let vars_from_formals = List.map variable_from_formal fn.a_formals in
	ignore (List.map (add_variable f_env.scope) vars_from_formals);

	(* Create new body thats typechecked *)
	let new_body = List.map (stmt f_env) fn.a_old_body in

	let new_function = {a_fname = fn.a_fname;
						a_is_stream = fn.a_is_stream;
						a_ret_type = fn.a_ret_type;
						a_formals = fn.a_formals;
						a_locals = f_env.scope.variables;
						a_old_body = [];
						a_body = new_body;
	} in
	print_endline "***Function Locals***";
	ignore (List.map string_of_variables new_function.a_locals);
	print_endline "==============Finished checking function==============";
	new_function

let string_of_functions fn =
	print_endline fn.a_fname

let string_of_scope scope =
	ignore (List.map string_of_variables scope.variables);
	List.map string_of_functions scope.functions
	
let print_env env =
	string_of_scope env.scope
	
let infer_prog program = 
	let vdecls, fdecls = program in
	let global_env = get_global_environment in
	let a_vdecls = List.map (check_vdecls global_env) vdecls in
	ignore (List.map (add_function global_env) fdecls);
	let a_fdecls = List.map (check_function_bodies global_env) global_env.scope.functions in
	
	print_endline "==============Printing Environment==============";
	ignore (print_env global_env);
	print_endline "==============Finished printing Environment==============";

	(a_vdecls, a_fdecls)
