%{ open Ast %}

%token SEMI LPAREN RPAREN LBRACE RBRACE COMMA COLON
%token LBRACKET RBRACKET
%token PLUS MINUS TIMES DIVIDE ASSIGN EXP
%token EQ NEQ LT LEQ GT GEQ
%token RETURN IF ELSEIF ELSE FOR WHILE INT FLOAT BOOL STRING
%token ARRAY 
%token SUM TO LET STREAM SIGNAL TIME PRINT
%token <bool> BOOLEAN_LIT
%token TIME
%token <float> F_LIT
%token <int> LITERAL
%token <string> ID
%token <string> STRING_LIT
%token EOF

%nonassoc NOELSE
%nonassoc ELSE
%nonassoc ELSEIF
%right ASSIGN
%left EQ NEQ
%left LT GT LEQ GEQ
%left PLUS MINUS
%left TIMES DIVIDE
%left EXP
%start program
%type <Ast.program> program

%%

program:
   /* nothing */ { [], [] }
 | program vdecl { ($2 :: fst $1), snd $1 }
 | program fdecl { fst $1, ($2 :: snd $1) }

fdecl:
	prim ID LPAREN formals_opt RPAREN LBRACE stmt_list RBRACE
     { { is_stream = false;
         fname = $2;
		 ret_type = Some($1);
	 	 formals = $4;
	 	 body = List.rev $7 } }
 
   | STREAM ID LPAREN formals_opt RPAREN LBRACE stmt_list RBRACE
     { { is_stream = true;
         fname = $2;
		 ret_type = None;
         formals = $4;
         body = List.rev $7; } }

formals_opt:
    /* nothing */ { [] }
  | formal_list   { List.rev $1 }

formal_list:
    formal              	 { [$1] }
  | formal_list COMMA formal { $3 :: $1 }

formal:
	prim ID
	{ {
		form_type	= $1;
		form_name	= $2;
	} }

variable:
	prim ID assign_opt SEMI
	{ {
		typ  = $1;
		name = $2;
		exp  = $3;
	} }

vdecl:
	variable	{ Variable_Dec($1) }

stmt_list:
    /* nothing */  { [] }
  | stmt_list stmt { $2 :: $1 }

stmt:
    expr SEMI { Expr($1) }
  | RETURN expr SEMI { Return($2) }
  | LBRACE stmt_list RBRACE { Block(List.rev $2) }
  | IF LPAREN expr RPAREN stmt %prec NOELSE { If($3, $5, Block([])) }
  | variable 	{ Prim_Assign($1) }
  | IF LPAREN expr RPAREN stmt ELSE stmt    { If($3, $5, $7) }
  | FOR LPAREN expr_opt SEMI expr_opt SEMI expr_opt RPAREN stmt
     { For($3, $5, $7, $9) }
  | WHILE LPAREN expr RPAREN stmt { While($3, $5) }
  | PRINT expr SEMI { Print($2) }
  | prim ID ASSIGN SUM ID ASSIGN LITERAL TO LITERAL COLON expr SEMI 	{ Sum($2, $5, $7, $9, $11) }
  | LET ID LBRACKET TIME RBRACKET ASSIGN SUM ID ASSIGN LITERAL TO LITERAL COLON expr SEMI 	{ Basic_Sum($2, $8, $10, $12, $14) }
  | create_basic SEMI	{ $1 }

create_basic:
  LET ID ASSIGN basic LBRACKET lit_opt RBRACKET { Basic_Dec($2, $4, $6) }

lit_opt:
	/* nothing */	{ -1 }
  | LITERAL			{ $1 }

expr_opt:
    /* nothing */ { Noexpr }
  | expr          { $1 }

expr:
    LITERAL          									{ Literal($1) }
  | STRING_LIT		 									{ String_literal($1) }
  | F_LIT			 									{ Float_literal($1) }
  | BOOLEAN_LIT										 	{ Bool_literal($1) }
  | ID assign_opt   									{ Id($1, $2) }
  | b_expr              								{ $1 }
  | ID LPAREN actuals_opt RPAREN						{ Call($1, $3) }
  | LPAREN expr RPAREN 									{ $2 }
  | basic_op											{ $1 }

b_expr:
    expr PLUS   expr 									{ Binop($1, Add,   $3) }
  | expr MINUS  expr 									{ Binop($1, Sub,   $3) }
  | expr TIMES  expr 									{ Binop($1, Mult,  $3) }
  | expr DIVIDE expr 									{ Binop($1, Div,   $3) }
  | expr EXP    expr 									{ Binop($1, Exp,   $3) }
  | expr EQ     expr 									{ Binop($1, Equal, $3) }
  | expr NEQ    expr 									{ Binop($1, Neq,   $3) }
  | expr LT     expr 									{ Binop($1, Less,  $3) }
  | expr LEQ    expr 									{ Binop($1, Leq,   $3) }
  | expr GT     expr 									{ Binop($1, Greater,  $3) }
  | expr GEQ    expr 									{ Binop($1, Geq,   $3) }

basic_op:
	ID LBRACKET LITERAL RBRACKET assign_opt			 { Basic_Op(Array, $1, $3, "!Noid", $5) }
  | ID LBRACKET ID RBRACKET assign_opt		{ Basic_Op(Array, $1, -1, $3, $5) }
  | ID LBRACKET time_expr RBRACKET assign_opt { Basic_Op(Signal, $1, $3, "!Noid", $5) }

time_expr:
	TIME							{ 0 }
  | TIME MINUS LITERAL				{ $3 }

assign_opt:
	/* nothing */ { Noexpr }
  | ASSIGN expr	  { $2 }

prim:
    STRING  { String }
  | INT    { Int }
  | FLOAT { Float }
  | BOOL   { Bool }

basic:
	ARRAY	{ Array }
  | SIGNAL	{ Signal }
     
actuals_opt:
    /* nothing */ { [] }
  | actuals_list  { List.rev $1 }

actuals_list:
    expr                    { [$1] }
  | actuals_list COMMA expr { $3 :: $1 }
